inherit "/classes/Script";
inherit "/kernel/package";

#include <macros.h>
#include <exception.h>
#include <classes.h>
#include <attributes.h>
#include <database.h>
#include <types.h>
#include <events.h>
#include <access.h>
#include <coal.h>

#define PACKAGE_NAME "web"
#define PACKAGE_VERSION "2.2.17"

#define XSL_T_CONTENT     "xsl:content"
#define XSL_T_ATTRIBUTES  "xsl:attributes"
#define XSL_T_ACCESS      "xsl:access"
#define XSL_T_ANNOTATIONS "xsl:details"
#define XSL_T_STYLESHEETS "xsl:stylesheets"

array get_mailbox_folders() {
  array ret = ({});
  object user = this_user();
  array(object) annos = user->get_annotations();
  foreach(annos, object anno) {
    if (objectp(anno) && (anno->get_object_class() & CLASS_CONTAINER) ) {
      ret += ({ anno });
    }
  }
  return ret;
}

/**
 * Show function to bring a config mapping in xml format.
 * It will have each configuration listed as config name='config-name'
 * with the value as container data.
 *
 * @param mapping configs - the configuration mapping to bring to xml.
 * @return the converted mapping.
 * @author <a href="mailto:astra@upb.de">Thomas Bopp</a>)
 */
string show_configs(mapping configs)
{
    string xml = "";
    foreach(indices(configs), mixed index) {
      if ( stringp(configs[index]) || intp(configs[index]) || floatp(configs[index]) )
         xml += "\t<config name=\""+index+"\"  changeable=\""+(_Server->is_config_changeable(index)?"true":"false")+"\">"+configs[index]+"</config>\n";
    }
    return xml;
}

/**
 * Process mailto attribute on messages
 */
string show_mailto(object|string|array mailto) {
  string ret = "";
  
  if (intp(mailto)) {
    ret = "";
  } else if (arrayp(mailto)) {
    int i = 0;
    foreach(mailto, object to) {
      if (stringp(to)) {
        if (i > 0) {
          ret += ", ";
        }
        ret += to;
        i++;
      }
    }
    // ret = _Server->get_module("Converter:XML")->compose(mailto);
  }
  else if (objectp(mailto)) {
   object converter = _Server->get_module("Converter:XML");
   ret += "<Object type="+converter->class_id_to_name(mailto->get_object_class())+">";
   ret += "<icon>";
   ret += converter->compose(mailto->query_attribute(OBJ_ICON));
   ret += "</icon>";
   ret += "<id>";
   ret += mailto->get_object_id();
   ret += "</id>";
   ret += "<description>";
   ret += mailto->query_attribute(OBJ_DESC);
   ret += "</description>";
   ret += "<name>";
   ret += mailto->query_attribute(OBJ_NAME);
   ret += "</name>";
   ret += "</Object>";
   //ret = _Server->get_module("Converter:XML")->compose(mailto);
  } else ret = "wrong mailto";
  return ret;
}


/**
 *  Get annotations recursive starting with the given object "anno"
 */
 array get_messages_recursive(object anno) {
//   werror("package:web: calling get_annos_rek anno="+anno->get_identifier()+"\n\r");
   if (!objectp(anno)) return ({ });
   array annos = anno->get_annotations();
   array rekannos = ({ });
   if (arrayp(annos)) {
     foreach(annos, object an) {
       if ( an->get_object_class() & CLASS_CONTAINER ) {
         rekannos += get_messages_recursive(an);
       }
       else rekannos += ({ an });
     }
   }
//   werror("rekannos=%O\r\n", rekannos);
   return rekannos;
 }

/**
 * Get the name of an object
 *
 * @param object obj - the document
 * @return the content of the document or "" for non-text documents.
 */
string get_message_name(string|object obj)
{ 
  if (stringp(obj)) {
      obj = find_object((int)obj);
  }
  if (!objectp(obj)) return "";
  return query_attribute(OBJ_NAME);
}
 
/**
 * Get the content of an object and detect the encoding. This only works
 * for text documents. If no utf8 encoding is detected it will be converted
 * to utf8.
 *
 * @param object obj - the document
 * @return the content of the document or "" for non-text documents.
 */
int get_message_id(string|object obj)
{
  if (stringp(obj)) {
      obj = find_object((int)obj);
  }
  if (!objectp(obj)) return -1;
  string mime = obj->query_attribute(DOC_MIME_TYPE);
  if (search(mime, "text/") != -1)
    return obj->get_object_id();
  
  array(object) annos = obj->get_annotations();
  foreach(annos, object anno) {
    mime = anno->query_attribute(DOC_MIME_TYPE);
    if (search(mime, "text/") != -1)
      return anno->get_object_id();
  }
  return -1;
}

/**
 * Get the content of an object and detect the encoding. This only works
 * for text documents. If no utf8 encoding is detected it will be converted
 * to utf8.
 *
 * @param object obj - the document
 * @return the content of the document or "" for non-text documents.
 */
string get_message_content(string|object obj)
{
  if (stringp(obj)) {
      obj = find_object((int)obj);
  }
  if (!objectp(obj)) return "";
  string mime = obj->query_attribute(DOC_MIME_TYPE);
  if (search(mime, "text/") != -1) 
    return get_annotation_content(obj);
  // mark message as read:
  obj->get_content();  // dont need to process result here
  
  array(object) annos = obj->get_annotations();
  foreach(annos, object anno) {
    mime = anno->query_attribute(DOC_MIME_TYPE);
    if (search(mime, "text/") != -1)
      return get_annotation_content(anno);
  }
  return "";
}

/**
 * Get the content of an object and detect the encoding. This only works
 * for text documents. If no utf8 encoding is detected it will be converted
 * to utf8.
 *
 * @param object obj - the document
 * @return the content of the document or "" for non-text documents.
 */
string get_annotation_content(string|object obj, void|object messageboard)
{
  if (stringp(obj)) {
      obj = find_object((int)obj);
  }
  if ( !objectp(obj) ) {
    if (objectp(messageboard)) {
      array annos = messageboard->get_annotations();
      if (arrayp(annos) && sizeof(annos) > 0) obj = annos[0];
    }
    if (!objectp(obj)) return "object not found";
  }
  if ( search(obj->query_attribute(DOC_MIME_TYPE), "txt") >= 0 ||
	 search(obj->query_attribute(DOC_MIME_TYPE), "text") >= 0 )
    {
	string content;
	string mtype = obj->query_attribute(DOC_MIME_TYPE);
	string encoding = obj->query_attribute(DOC_ENCODING);
	
	if ( stringp(mtype) && search(mtype, "wiki") >= 0 ) {
	  object wiki = get_module("wiki");
	  if ( objectp(wiki) ) {
	    content = wiki->wiki_to_html_plain(obj);
	    encoding = "utf-8";
	  }
	}
  else if ( stringp(mtype) && mtype == "text/plain" ) {
    content = obj->get_content();
    content = text_to_html(content);
  }
	else 
	  content = obj->get_content();

	if ( !stringp(encoding) &&
	     search(mtype, "html") >= 0 )
	    encoding = detect_encoding(content);

	if ( !stringp(content) )
	  content = "";
	if ( encoding != "utf-8" ) {
	  return "<![CDATA["+string_to_utf8(content)+"]]>";
	}
	else
	  return "<![CDATA["+content+"]]>";
    }
    return "";
}

void check_xsl_attr(string key, object comp, object obj, int oclass)
{
    mapping m = obj->query_attribute(key);
    if ( !mappingp(m) || m[_STEAMUSER] != comp )
    {
	MESSAGE("Re-registering "+key+" upon upgrading web package");
	object f = _Server->get_factory(oclass);
	f->re_register_attribute(key, REGISTERED_DEFAULT, comp);
	pck_log("Attribute: "+key + " faulty registration !\n");
    }
    else
	pck_log("Attribute: "+key + " OK!\n");
}

/**
 * Make sure everything is set correctly.
 *  
 * @author <a href="mailto:astra@upb.de">Thomas Bopp</a>) 
 */
bool
check_package_integrity()
{
  mapping comps = get_components();
  bool integrity = true;
	if ( mappingp(comps) ) {
    foreach(indices(comps), string comp) {
	string   fname, path;

	fname = basename(comp);
	path  = dirname(comp);
	
	object o = _FILEPATH->path_to_object(comp);
	if ( !objectp(o) ) {
	    pck_log("Component: " + comp + " not available at position '"+
		    comp+"'.\n");
	}
	object obj = comps[comp];
	if ( !objectp(obj) ) {
	    pck_log("Component: " + comp + " is no objectp!\n");
	    integrity = false;
	}
	else {
	    if ( obj->get_identifier() == "" ) {
		obj->set_attribute(OBJ_NAME, fname);
		object cont = get_component(path);
		if ( !objectp(cont) ) {
		    pck_log("Path component: " + path + " no available!\n");
		}
		else
		    obj->move(cont);
	    }
	}
    }
    }
    if ( !integrity )
	return false;
    mixed err = catch {
	// check for content.xsl etc.
	check_xsl_attr("xsl:content", 
		       get_component("/stylesheets/content.xsl"),
		       _ROOTROOM, CLASS_OBJECT);
	check_xsl_attr("xsl:content", 
		       get_component("/stylesheets/content.xsl"),
		       _ROOT, CLASS_OBJECT);
	
	check_xsl_attr("xsl:attributes", 
		       get_component("/stylesheets/attributes.xsl"),
		       _ROOTROOM, CLASS_OBJECT);
	check_xsl_attr("xsl:attributes", 
		       get_component("/stylesheets/attributes.xsl"),
		       _ROOT, CLASS_OBJECT);
	
	check_xsl_attr("xsl:access", 
		       get_component("/stylesheets/access.xsl"),
		       _ROOTROOM, CLASS_OBJECT);
	check_xsl_attr("xsl:access", 
		       get_component("/stylesheets/access.xsl"),
		       _ROOT, CLASS_OBJECT);
	
	check_xsl_attr("xsl:details", 
		       get_component("/stylesheets/user_details.xsl"),
		       _ROOT, CLASS_USER);
	
	check_xsl_attr("xsl:content", 
		       get_component("/stylesheets/group.xsl"),
		       _STEAMUSER, CLASS_GROUP);

	check_xsl_attr("xsl:content", 
		       get_component("/stylesheets/content.xsl"),
		       _ROOTROOM, CLASS_MESSAGEBOARD);

	set_urls();
  prepare_specials();
	object factory = get_factory(CLASS_CALENDAR);
	foreach(GROUP("steam")->get_members(), object mem) {
	    if ( !objectp(mem) )
		return 0;
	    object calendar = mem->query_attribute(USER_CALENDAR);
	    if ( !objectp(calendar) ) {
		calendar = factory->execute( (["name":mem->get_identifier()+"'s Calendar",]));
    mem->unlock_attribute(USER_CALENDAR);
		mem->set_attribute(USER_CALENDAR, calendar);
		mem->lock_attribute(USER_CALENDAR);
		catch(calendar->set_creator(mem));
    calendar->unlock_attribute(CALENDAR_OWNER);
		calendar->set_attribute(CALENDAR_OWNER, mem);
		calendar->lock_attribute(CALENDAR_OWNER);
		calendar->sanction_object(mem, SANCTION_ALL);
	    }
	}
	foreach(get_module("groups")->get_groups(), object grp) {
	    object calendar = grp->query_attribute(GROUP_CALENDAR);
	    if ( !objectp(calendar) ) {
		calendar = factory->execute( (["name":grp->get_identifier()+"'s Calendar",]));
		grp->set_attribute(GROUP_CALENDAR, calendar);
		grp->lock_attribute(GROUP_CALENDAR);
		calendar->set_attribute(CALENDAR_OWNER, grp);
		calendar->sanction_object(grp, SANCTION_ALL);
	    }
	}
    };

    if ( err != 0 ) {
	FATAL("Error while trying to fix package integrity - reinstalling!");
	FATAL(sprintf("%O\n",err));
	return false;
    }
    
    return true;
}

static void low_set_url(object obj, string url)
{
 object old = _Server->get_module("filepath:url")->path_to_object( url );
 if (objectp(old)) {
   old->set_attribute(OBJ_URL, 0);
 }
 mixed err = catch {
   obj->set_attribute(OBJ_URL, url);
   obj->sanction_object(_WORLDUSER, SANCTION_READ);
 };
 if ( err != 0 )
   FATAL("Error while publishing object %O\n%O: %O", obj, err[0], err[1]);
}          

void set_urls()
{
  // obsolete ?  see setup.xml (needs test)
    low_set_url(OBJ("/scripts"), "/scripts");
    low_set_url(OBJ("/images"),  "/images");
    low_set_url(OBJ("/gui"), "/gui");
    low_set_url(OBJ("/gui_css"), "/gui_css");
    low_set_url(OBJ("/gui_js"), "/gui_js");
    low_set_url(OBJ("/styles"), "/styles");
    low_set_url(OBJ("/backtraces"), "/backtraces");
    low_set_url(OBJ("/register"), "/register");
    low_set_url(OBJ("/stylesheets"), "/stylesheets");
    low_set_url(OBJ("/documents"), "/documents");
    low_set_url(OBJ("/languages"), "/languages");
}

void prepare_specials() {
  // delete things which are no longer needed to clean up WI code
  object del = OBJ("/sources/testscorm.pike");
  if (objectp(del)) del->delete();
  del = OBJ("/sources/createscorm.pike");
  if (objectp(del)) del->delete();
  del = OBJ("/documents/create.xml");
  if (objectp(del)) del->delete();  
  
  // only publish index.xml if web package wasnt installed before
  // this helps keeping settings if e.g. chinese_registration was installed
  // or admin modified "/" (e.g. on gauge.upb.de)
  if (!objectp(_Server->get_module("package:web"))) {
    object prev=_Server->get_module("filepath:url")->resolve_path(0,"/");
    if ( objectp(prev) ) {
      string stri ="/ was occupied by object nr"+ prev->get_object_id() + " name="+prev->get_identifier() + "  Registering index.xml for / anyway";
      LOG( stri );
      prev->set_attribute(OBJ_URL, 0);
    }
  }

  OBJ("/index.xml")->sanction_object(_WORLDUSER, SANCTION_READ|SANCTION_EXECUTE|SANCTION_ANNOTATE);
  OBJ("/index.xml")->set_attribute( "xsl:document" , OBJ("/stylesheets/start.xsl"));
  OBJ("/index.xml")->set_attribute(OBJ_URL, "/");

  OBJ("/documents/access.xml")->set_attribute("xsl:document", OBJ("/stylesheets/errors.xsl"));
  
  // setting stylesheet for (technical) changed create process
  object mycreate = OBJ("/documents/create");
  if (objectp(mycreate)) {
    mapping mymaps = mycreate->query_attribute("xsl:content");
    if (!mappingp(mymaps)) mymaps = ([ ]);
    mymaps[GROUP("steam")] = OBJ("/stylesheets/create.xsl");
    mycreate->set_attribute( "xsl:content" , mymaps );
  }

  object mapview = OBJ("/tools/mapview.pike");
  if (objectp(mapview)) {
    mapview->provide_instance();
  }
  
  // checking root-room stylesheet settings
  // if stylesheet of rootroom was set to the redirect of the spm_support
  // module delete stylesheet setting to use /stylesheets/content.xsl
  object rootroom = OBJ("/");
  mapping settings = rootroom->query_attribute("xsl:content");
  if (mappingp(settings)) {
    object stylesheet = settings[GROUP("steam")];
    if (objectp(stylesheet)) {
      object s2 = OBJ("/packages/spm_support/stylesheets/redirect.xsl");
      if (objectp(s2) && stylesheet == s2) {
        rootroom->set_attribute("xsl:content", 0);
      }
    }
  }
  // check if groups.pike is available as /groups
  if ( !objectp(OBJ("/groups")) ) 
  {
    object group_script = OBJ("/sources/groups.pike")->execute( ([ "name":"groups",]));
    group_script->move(_ROOTROOM);
    group_script->sanction_object(_STEAMUSER, SANCTION_EXECUTE|SANCTION_READ);
  }
  // check if calendar.pike is available as /calendars
  if ( !objectp(OBJ("/calendars")) ) 
  {
    object cal = OBJ("/sources/calendar.pike")->execute( ([ "name":"calendars",]));
    cal->move(_ROOTROOM);
    cal->sanction_object(_STEAMUSER, SANCTION_EXECUTE|SANCTION_READ);
  }
  // check if spm-module is available as /SPM
  object spm = OBJ("/SPM");
  
  if ( !objectp( spm )) {
    spm = _Server->get_module("SPM");
    if (objectp(spm)) {
      spm->move(OBJ("/"));
    }
  }
  if (objectp( spm ) ) {
    spm->set_attribute("OBJ_NAME", "SPM");
    spm->set_attribute("OBJ_DESC", "steam package manager module");
    spm->set_acquire_attribute(OBJ_ICON, _Server->get_module("icons"));
  }
}

array(object) install_scripts()
{
    object script_folder = _FILEPATH->path_to_object("/scripts");
    object source_folder = get_component("/sources");
    array(object) scripts = ({ });

    if ( !objectp(source_folder) ) {
	source_folder = _FILEPATH->path_to_object("/sources");
	if ( !objectp(source_folder) ) 
	    THROW("No sources found - aborting !", E_ERROR);
    }
    source_folder->sanction_object(_STEAMUSER, SANCTION_READ|SANCTION_EXECUTE);

    if (!objectp(script_folder)) script_folder = _FILEPATH->make_directory("/", "scripts");
    foreach ( source_folder->get_inventory(), object source ) {
      if ( source->get_object_class() & CLASS_DOCLPC ) {
        
        string sname = source->get_identifier();
        object oinstance = script_folder->get_object_byname(sname);

        if (!objectp( oinstance )) {
          mixed err = catch {
            object script = 
            source->execute( ([ "name": sname, ]) );
            scripts += ({ script });
            script->move(script_folder);
            check_name(script);
          };
        }
      }
    }
    script_folder->sanction_object(_STEAMUSER, SANCTION_EXECUTE|SANCTION_READ);
    return scripts;
}

void check_name(object script)
{ 
    string newname = 0;
    switch ( script->get_identifier() ) {
         case "export.pike":
           newname = "export.tar";
	     break;
         case "svg.pike":
	         newname = "svg.svg";
                break;
         case "tar.pike":
	         newname = "tar.tar";
                break;
         case "sal.pike":
           newname = "steam.sal";
                break;
    }
    if (stringp( newname )) {
      // check if object with the required name already exists and delete it
      object old = OBJ("/scripts/" + newname);
      if (objectp(old)) old->delete();
      // rename the instance in the scrips - folder to the required new name
      script->set_attribute(OBJ_NAME, newname);
    }
}

void reload_xml_structures() {
  object stylesheet_folder = _FILEPATH->path_to_object("/stylesheets");
  array stylesheets = stylesheet_folder->get_inventory_by_class(CLASS_DOCXSL);
  foreach (stylesheets, object xsl) {
    if (objectp(xsl)) {
      mixed err = catch( xsl->load_xml_structure() );
      if (err) FATAL("web.pike: Failed to load XML Structure of %O\n%O:%O", xsl, err[0], err[1]);
    }
  }
}

array upgrade_scripts()
{
    array scripts = ({ });

    object script_folder = _FILEPATH->path_to_object("/scripts");
    if ( !objectp(script_folder) ) {
	THROW("Failed to find script folder - aborting...", E_ERROR);
    }
    object source_folder = get_component("/sources");
    if ( !objectp(source_folder) ) {
      source_folder = _FILEPATH->path_to_object("/sources");
      if ( !objectp(source_folder) ) 
          THROW("No sources found - aborting !", E_ERROR);
    }
    foreach ( source_folder->get_inventory(), object source ) {
        
        if ( source->get_object_class() & CLASS_DOCLPC ) {
        array instances = source->get_instances();
	    mixed err = catch {
		source->upgrade();
		foreach( instances, object src ) {
		    if ( src->get_environment() == script_folder )
			scripts += ({ src });
		}
	    };
            if ( sizeof(source->get_instances()) == 0 ) {
               err = catch {
                    object script = source->execute
		      ( ([ "name": source->get_identifier(), ]) );
                    script->move(script_folder);
                    check_name(script);
               }; 
               if ( err != 0 )
		   werror("Error creating script:"+sprintf("%O\n", err));
            }
	    else {
		foreach(instances, object inst) {
		    if ( !objectp(inst) )
		      continue;
		    if ( !stringp(inst->get_identifier()) )
                        inst->delete();
                    else
		        check_name(inst);
                }
		instances = source->get_instances();
		if ( sizeof(instances) == 1 ) {
		    object inst = source->get_instances()[0];
		    if ( inst->get_environment() != script_folder ) 
			inst->move(script_folder);
		}
	    }	
	}
	else {
	  werror("Stale file in source directory " + source->get_identifier() + "\n");
	}
    }
    return scripts;
}

void install_tags()
{
  object tagfolder = OBJ("/tags");
  if ( !objectp(tagfolder) )
    steam_error("Failed to find tag folder /tags - aborting installation !");
  
  array tags = tagfolder->get_inventory();
  foreach( tags, object tag ) {
    if ( tag->get_object_class() & CLASS_DOCLPC ) {
      object instance = tag->get_instance();
      if ( !objectp(instance) ) {
	tag->execute( (["name":htmllib.get_tag_name(tag),]) );
      }
    }
  }
  tagfolder->sanction_object( _WORLDUSER, SANCTION_READ|SANCTION_EXECUTE);
}

array spm_upgrade_package()
{
  // doing some integrity stuff, checking for /calendars /groups and some more...
  set_urls();
  prepare_specials();
  
  // provide attributes as needed by webinterface
    my_provide_attributes();
  
    // upgrade all scripts of the package
    _Server->get_module("icons")->init_icons();
    install_tags();
//    catch(OBJ("/chat")->set_attribute("url", "/chat"));
    get_module("filepath:url")->set_attribute("xsl:content", 
			    ([ GROUP("steam"): OBJ("/stylesheets/url.xsl"), ]));
    OBJ("/documents/errors")->set_attribute("language_index", ([ 
      "german": "german.html",
      "english": "english.html",
      "chinese": "chinese.html",
    ]));
    do_set_attribute(PACKAGE_VERSION, get_version());
    
    if ( objectp( OBJ("/wiki") )) {
        OBJ("/wiki")->set_acquire_attribute(OBJ_ICON, _Server->get_module("icons"));
    }

    if ( objectp( OBJ("/home") )) {
        OBJ("/home")->set_acquire_attribute(OBJ_ICON, _Server->get_module("icons"));
    }

    if ( objectp( OBJ("/calendar") )) {
        OBJ("/calendar")->set_acquire_attribute(OBJ_ICON, _Server->get_module("icons"));
    }
    
    reload_xml_structures();
    
    return upgrade_scripts();
}

void my_provide_attributes() {
    object steam = _STEAMUSER;
    
    provide_attribute(CLASS_OBJECT, OBJ_URL, CMD_TYPE_STRING, 
		      "url",
		      0, EVENT_ATTRIBUTES_CHANGE, get_module("filepath:url"),
		      CONTROL_ATTR_USER, 0);
    provide_attribute(CLASS_OBJECT, "xsl:content", CMD_TYPE_MAPPING, 
		      "xsl:content",
		      0, EVENT_ATTRIBUTES_CHANGE, REG_ACQ_ENVIRONMENT,
		      CONTROL_ATTR_USER, 
		      ([ steam:OBJ("/stylesheets/content.xsl"), ]),
		      _ROOTROOM);
    provide_attribute(CLASS_OBJECT, "xsl:attributes", 
		      CMD_TYPE_MAPPING, "xsl:attributes", 
		      0, EVENT_ATTRIBUTES_CHANGE, REG_ACQ_ENVIRONMENT,
		      CONTROL_ATTR_USER, 
		      ([steam:OBJ("/stylesheets/attributes.xsl"),]),
		      _ROOTROOM);
    provide_attribute(CLASS_OBJECT, "xsl:access", CMD_TYPE_MAPPING, 
		      "xsl:access",
		      0, EVENT_ATTRIBUTES_CHANGE, REG_ACQ_ENVIRONMENT,
		      CONTROL_ATTR_USER, 
		      ([ steam:OBJ("/stylesheets/access.xsl"), ]),
		      _ROOTROOM);
    provide_attribute(CLASS_OBJECT, "xsl:public", CMD_TYPE_OBJECT, 
		      "xsl:public",
		      0, EVENT_ATTRIBUTES_CHANGE, REG_ACQ_ENVIRONMENT,
		      CONTROL_ATTR_USER, 
		      OBJ("/stylesheets/public.xsl"), 0);
    provide_attribute(CLASS_OBJECT, "xsl:annotations", 
		      CMD_TYPE_MAPPING, "xsl:annotations",
		      0, EVENT_ATTRIBUTES_CHANGE, REG_ACQ_ENVIRONMENT,
		      CONTROL_ATTR_USER,
		      ([steam:OBJ("/stylesheets/annotations.xsl"),]),
		      _ROOTROOM);
    provide_attribute(CLASS_USER,"xsl:access", CMD_TYPE_MAPPING, "xsl:access",
		      0, EVENT_ATTRIBUTES_CHANGE, _ROOT,
		      CONTROL_ATTR_USER,
		      ([steam:OBJ("/stylesheets/access.xsl"),]),
		      _ROOT);
    provide_attribute(CLASS_USER,"xsl:content", CMD_TYPE_MAPPING, 
		      "xsl:content",
		      0, EVENT_ATTRIBUTES_CHANGE, _ROOT, CONTROL_ATTR_USER, 
		      ([ steam:OBJ("/stylesheets/content.xsl"), ]),
		      _ROOT);
    provide_attribute(CLASS_USER,"xsl:details", CMD_TYPE_MAPPING, 
		      "xsl:details", 0, EVENT_ATTRIBUTES_CHANGE, _ROOT,
		      CONTROL_ATTR_USER,([
			  steam:OBJ("/stylesheets/user_details.xsl"),
			  ]), _ROOT);
    provide_attribute(CLASS_GROUP,"xsl:access", CMD_TYPE_MAPPING,"xsl:access",
		      0, EVENT_ATTRIBUTES_CHANGE, _WORLDUSER,CONTROL_ATTR_USER,
		      ([ steam:OBJ("/stylesheets/access.xsl"), ]),
		      _WORLDUSER);
    provide_attribute(CLASS_GROUP,"xsl:content", CMD_TYPE_MAPPING,
		      "xsl:content", 0, EVENT_ATTRIBUTES_CHANGE, _WORLDUSER,
		      CONTROL_ATTR_USER, 
		      ([ steam:OBJ("/stylesheets/group.xsl"), ]),
		      _WORLDUSER);
    if ( get_factory(CLASS_USER)->describe_attribute("user_selection") )
      get_factory(CLASS_USER)->unregister_attribute("user_selection");
    provide_attribute(CLASS_USER, "USER_SELECTION", CMD_TYPE_ARRAY,
		      "user-selection", 
		      EVENT_ATTRIBUTES_QUERY, EVENT_ATTRIBUTES_CHANGE, 0,
		      CONTROL_ATTR_CLIENT, ({ }) );
    provide_attribute(CLASS_DOCUMENT, "xsl:document", CMD_TYPE_OBJECT,
		      "xsl:document",
		      0, EVENT_ATTRIBUTES_CHANGE, REG_ACQ_ENVIRONMENT,
		      CONTROL_ATTR_USER, 0);
    provide_attribute(CLASS_CONTAINER, "web:sort:exits", CMD_TYPE_STRING,
		      "sort exits",
		      0, EVENT_ATTRIBUTES_CHANGE, 0, CONTROL_ATTR_USER,
		      "last-modified/time", 0);
    provide_attribute(CLASS_CONTAINER, "web:sort:objects", CMD_TYPE_STRING,
                      "sort objects",
		      0, EVENT_ATTRIBUTES_CHANGE, 0, CONTROL_ATTR_USER,
		      "name", 0);
}

void install_package(string source, void|string version)
{
}

array spm_install_package()
{
    array(object) scripts = install_scripts();
    object steam = _STEAMUSER;

    my_provide_attributes();
          
    object groupmodule = _Server->get_module("groups");
    if (objectp(groupmodule)) {
      array groups = groupmodule->get_groups();
      foreach(groups, object group) {
        if (objectp(group)) {
          if ( group->get_group_name() != "Everyone" )
            group->set_acquire_attribute("xsl:content", GROUP("Everyone"));
        }
      }
    }
    object factory = _Server->get_factory(CLASS_USER);
    factory = factory->this();
    
    object svg = OBJ("/scripts/svg.pike");
    if ( objectp(svg) ) {
      object old_svg = OBJ("/scripts/svg.svg");
      if ( objectp(old_svg) ) old_svg->delete();
	svg->set_attribute(OBJ_NAME, "svg.svg");
    }
    if ( objectp(OBJ("/scripts/export.pike")) ) {
      object old_export = OBJ("/scripts/export.tar");
      if ( objectp(old_export) ) old_export->delete();
      OBJ("/scripts/export.pike")->set_attribute(OBJ_NAME, "export.tar");
    }
    object sal = OBJ("/scripts/sal.pike");
    if ( objectp(sal) ) {
      object old_sal = OBJ("/scripts/sal.pike");
      if ( objectp(old_sal) ) old_sal->delete();
	sal->set_attribute(OBJ_NAME, "steam.sal");
    }
    mixed err = catch {
	_Server->get_module("icons")->init_icons();
    };
    set_urls();
    prepare_specials();
    OBJ("/documents/errors")->set_attribute("language_index", ([ 
      "german": "german.html",
      "english": "english.html",
      "chinese": "chinese.html",
    ]));
    get_module("filepath:url")->set_attribute("xsl:content", 
      ([ GROUP("steam"): OBJ("/stylesheets/url.xsl"), ]));
      
    OBJ("/packages")->set_attribute("xsl:content", 
      ([ GROUP("steam"): OBJ("/stylesheets/spm.xsl"), ]));
      
    if ( !objectp( OBJ("/SPM") )) {
      object spm = _Server->get_module("SPM");
      if (objectp( spm ) ) {
        spm->move(OBJ("/"));
        spm->set_attribute("OBJ_NAME", "SPM");
        spm->set_attribute("OBJ_DESC", "steam package manager module");
        spm->set_acquire_attribute(OBJ_ICON, _Server->get_module("icons"));
      }
    }

    if ( objectp( OBJ("/wiki") )) {
        OBJ("/wiki")->set_acquire_attribute(OBJ_ICON, _Server->get_module("icons"));
    }

    if ( objectp( OBJ("/home") )) {
        OBJ("/home")->set_acquire_attribute(OBJ_ICON, _Server->get_module("icons"));
    }

    if ( objectp( OBJ("/calendar") )) {
        OBJ("/calendar")->set_acquire_attribute(OBJ_ICON, _Server->get_module("icons"));
    }

    
    install_tags();
    do_set_attribute(PACKAGE_VERSION, get_version());
    require_save();
    
    reload_xml_structures();
    
    return scripts;
}


/**
 * Callback function upon installation adds a global event to
 * event registration. 
 *  
 */
static void install_module()
{    
    add_global_event(EVENT_REGISTER_MODULE, event_register_module, 
		     PHASE_NOTIFY);

    // make sure the folders are in the rootroom if they still exist
    array(string) folders = ({ "/scripts", "/sources", "/gui_js",
			       "/register", "/stylesheets" });
    object obj, comp;

    foreach ( folders, string folder ) {
	obj = _FILEPATH->path_to_object(folder);
	comp = get_component(folder);
	if ( !objectp(comp) ) {
	    pck_log("Component: "+ folder + " is not available...\n");
	    if ( objectp(obj) )
		add_component(folder, obj);
	}
	if ( (!objectp(obj) || obj != comp) && objectp(comp) ) {
	    LOG("Moving folder to new rootroom !\n");
	    comp->move(_ROOTROOM);
	}
    }
}

static void load_stylesheets()
{
    if ( !get_factory("Object") )
	call_out(load_stylesheets, 1);
    array stylesheets = OBJ("/stylesheets")->
	get_inventory_by_class(CLASS_DOCXSL);
    call_out(load_stylesheet, 0.1, stylesheets, 0);
}

void load_stylesheet(array stylesheets, int num)
{
    object s = stylesheets[num];
    catch {
      s->get_stylesheet("english");
      s->get_stylesheet("german");
      s->get_stylesheet("chinese");
    };
    num++;
    if ( num < sizeof(stylesheets) )
      call_out(load_stylesheet, 0.2, stylesheets, num);
}

void load_module()
{
  call_out(load_stylesheets, 0);
}

/**
 * Called when a new module is registered.
 *  
 * @param object caller - the calling object
 * @param object module - the newly registered module.
 * @author Thomas Bopp (astra@upb.de) 
 */
void event_register_module(object caller, object module)
{
}

/**
 * Set a mapping of files which adds components to the package.
 *  
 * @param mapping files - the file mapping to add as components.
 * @author Thomas Bopp (astra@upb.de) 
 * @see add_component
 */
void set_files(mapping files)
{
    LOG("Setting files...\n");
    if ( _SECURITY->access_write(this(), CALLER) ) {
	foreach(indices(files), string file) {
	    add_component(file, this());
	}
    }
}
 
/**
 * Set the configurations of the server. After installation the 
 * web package sets some vars like webserver and ports.
 *  
 * @param mapping cfgs - the configurations added to the server.
 * @author Thomas Bopp (astra@upb.de) 
 */
void set_configs(mapping cfgs)
{
    if ( _SECURITY->access_write(this(), CALLER) ) {
	foreach ( indices(cfgs), string cfg ) {
	    if ( strlen(cfg) > 3 && cfg[0..3] == "web:" )
		_Server->set_config(cfg, cfgs[cfg]);
	}
    }
}

string get_identifier() { return "package:web"; }
string get_version() { return PACKAGE_VERSION; }


/**
 * Called after installation to make sure everything was installed fine.
 *  
 * @return true or false.
 * @author Thomas Bopp (astra@upb.de) 
 */
static bool validate_installation()
{
    return true;
}


